package mv3302;

import java.awt.geom.Point2D;
import static java.lang.Math.sqrt;
import java.util.ArrayList;
import java.util.List;
import simkit.SimEntityBase;
import static simkit.smd.Math2D.ZERO;
import simkit.smd.Mover;
import simkit.smd.Sensor;
/**
The SimpleReferee class represents a referee in a simulation. It extends the SimEntityBase class,
which provides basic functionality for simulation entities.
*/
public final class SimpleReferee extends SimEntityBase {

    private List<Sensor> sensors;

    private List<Mover> targets;

    /**
Constructs a SimpleReferee object with an empty list of sensors and targets.
*/
    
    public SimpleReferee() {
    }

    /**
Constructs a SimpleReferee object with the specified lists of sensors and targets.
@param sensors the list of sensors to set
@param targets the list of targets to set
*/
    
    public SimpleReferee(List<Sensor> sensors, List<Mover> targets) {
        setSensors(sensors);
        setTargets(targets);
    }

    /**
Performs the run logic for the referee. If there are sensors present, it waits for the "InitSensor"
event with zero delay.
*/
    
    public void doRun() {
        if (!sensors.isEmpty()) {
            waitDelay("InitSensor", 0.0, 0);
        }
    }

    /**
Initializes a sensor at the specified index. Adds the sensor as a simulation event listener.
If there are more sensors to initialize, it waits for the "InitSensor" event for the next index.
@param i the index of the sensor to initialize
*/
    
    public void doInitSensor(int i) {
        sensors.get(i).addSimEventListener(this);

        if (i < sensors.size() - 1) {
            waitDelay("InitSensor", 0.0, i + 1);
        }

        waitDelay("InitTarget", 0.0, i, 0);
    }

    /**
Initializes a target at the specified indices. Adds the target as a simulation event listener.
If there are more targets to initialize, it waits for the "InitTarget" event for the next indices.
Checks if the target is within range of the sensor and waits for the "EnterRange" event if necessary.
@param i the index of the sensor
@param j the index of the target
*/
    
    public void doInitTarget(int i, int j) {
        targets.get(j).addSimEventListener(this);
        if (j < targets.size() - 1) {
            waitDelay("InitTarget", 0.0, i, j + 1);
        }

        if (targets.get(j).getCurrentLocation().distance(sensors.get(i).getCurrentLocation()) < sensors.get(i).getMaxRange()) {
            if (!targets.get(j).equals(sensors.get(i).getMover())) {
                waitDelay("EnterRange", 0.0, targets.get(j), sensors.get(i));
            }
        }
    }
  
    /**
Updates a sensor with a target. Checks if the sensor and target are different and calculates
the enter and exit times. Handles different cases based on the number of times calculated.
If there are more sensors to update, it waits for the "UpdateSensor" event for the next index.
@param i the index of the sensor
@param target the target to update the sensor with
*/
    
    public void doUpdateSensor(int i, Mover target) {
        Sensor sensor = sensors.get(i);
        if (!sensor.equals(target)) {
            double[] times = getEnterExitTimes(target, sensor);
            switch (times.length) {
                case 0 -> {
                }
                case 1 -> {
                    interrupt("ExitRange", target, sensor);
                    waitDelay("ExitRange", times[0], target, sensor);
                }
                case 2 -> {
                    interrupt("EnterRange", target, sensor);
                    waitDelay("EnterRange", times[0], target, sensor);
                }
            }
        }

        if (i < sensors.size() - 1) {
            waitDelay("UpdateSensor", 0.0, i + 1, target);
        }
    }
    
    /**
 * Notifies the referee that a target is starting to move. Initiates the "UpdateSensor" event
 * with zero delay for the specified target.
 *
 * @param target the target that is starting to move
 */
    
    public void doStartMove(Mover target) {
        waitDelay("UpdateSensor", 0.0, 0, target);
    }

   /**
 * Notifies the referee that a target has stopped. Initiates the "UpdateSensor" event
 * with zero delay for the specified target.
 *
 * @param target the target that has stopped
 */ 
    
    public void doStop(Mover target) {
        waitDelay("UpdateSensor", 0.0, 0, target);
    }

    /**
 * Updates a target with a sensor. Checks if the target and sensor are different and calculates
 * the enter and exit times. Handles different cases based on the number of times calculated.
 * If there are more targets to update, it waits for the "UpdateTarget" event for the next index.
 *
 * @param i      the index of the target
 * @param sensor the sensor to update the target with
 */
    
    public void doUpdateTarget(int i, Sensor sensor) {
        Mover target = targets.get(i);
        double[] times = getEnterExitTimes(target, sensor);
        switch (times.length) {
            case 0 -> {
            }
            case 1 -> {
                interrupt("ExitRange", target, sensor);
                waitDelay("ExitRange", times[0], target, sensor);
            }
            case 2 -> {
                interrupt("EnterRange", target, sensor);
                waitDelay("EnterRange", times[0], target, sensor);
            }
        }
        
        if (i < targets.size() - 1) {
            waitDelay("UpdateTarget", i + 1, sensor);
        }
    }

    /**
 * Notifies the referee that a sensor is starting to move. Initiates the "UpdateTarget" event
 * with zero delay for the specified sensor.
 *
 * @param sensor the sensor that is starting to move
 */
    
    public void doStartMove(Sensor sensor) {
        waitDelay("UpdateTarget", 0.0, 0, sensor);
    }

    /**
 * Notifies the referee that a sensor has stopped. Initiates the "UpdateTarget" event
 * with zero delay for the specified sensor.
 *
 * @param sensor the sensor that has stopped
 */
    
    public void doStop(Sensor sensor) {
        waitDelay("UpdateTarget", 0.0, 0, sensor);
    }
    
    /**
 * Performs the actions when a target enters the range of a sensor. Calculates the enter and exit times
 * and waits for the "ExitRange" event if necessary.
 *
 * @param target the target that entered the range
 * @param sensor the sensor that detected the target
 */

    public void doEnterRange(Mover target, Sensor sensor) {
        double[] times = getEnterExitTimes(target, sensor);
        if (times.length == 1) {
            waitDelay("ExitRange", times[0], target, sensor);
        } else if (times.length == 2) {
            waitDelay("ExitRange", times[1], target, sensor);
        }
    }
    
    /**
 * Performs the actions when a target exits the range of a sensor.
 *
 * @param target the target that exited the range
 * @param sensor the sensor that lost track of the target
 */
    
    public void doExitRange(Mover target, Sensor sensor) {
    }

    /**
 * Calculates the enter and exit times for a target and a sensor based on their relative locations
 * and velocities. Returns an array of the calculated times.
 *
 * @param target the target to calculate times for
 * @param sensor the sensor to calculate times for
 * @return an array of the enter and exit times, or an empty array if no times were calculated
 */
    
    private double[] getEnterExitTimes(Mover target, Sensor sensor) {
        double[] times;

        Point2D relativeLocation = new Point2D.Double(
                target.getCurrentLocation().getX() - sensor.getCurrentLocation().getX(),
                target.getCurrentLocation().getY() - sensor.getCurrentLocation().getY()
        );
        Point2D relativeVelocity = new Point2D.Double(
                target.getVelocity().getX() - sensor.getVelocity().getX(),
                target.getVelocity().getY() - sensor.getVelocity().getY()
        );

        double velocitySquared = relativeVelocity.distanceSq(ZERO);
        double distanceSqured = relativeLocation.distanceSq(ZERO);

        double innerProduct = relativeLocation.getX() * relativeVelocity.getX()
                + relativeLocation.getY() * relativeVelocity.getY();

        double determinant = velocitySquared * (sensor.getMaxRange() * sensor.getMaxRange()
                - distanceSqured) + innerProduct * innerProduct;
        if (determinant < 0.0) {
            times = new double[0];
        } else {
            double first = -innerProduct / velocitySquared;
            double smallerRoot = first - sqrt(determinant) / velocitySquared;
            double biggerRoot = first + sqrt(determinant) / velocitySquared;
            if (smallerRoot > 0.0) {
                times = new double[]{smallerRoot, biggerRoot};
            } else if (biggerRoot > 0) {
                times = new double[]{biggerRoot};
            } else {
                times = new double[0];
            }
        }

        return times;
    }

/**
 * Returns a new list containing the sensors in the simulation.
 * @return a list of sensors
 */
    public List<Sensor> getSensors() {
        return new ArrayList<>(sensors);
    }

/**
 * Sets the list of sensors for the simulation.
 *
 * @param sensors the list of sensors to set
 */
    public void setSensors(List<Sensor> sensors) {
        this.sensors = new ArrayList<>(sensors);
    }

/**
 * Returns a new list containing the targets in the simulation.
 *
 * @return a list of targets
 */
    public List<Mover> getTargets() {
        return new ArrayList<>(targets);
    }

/**
 * Sets the list of targets for the simulation.
 *
 * @param targets the list of targets to set
 */
    public void setTargets(List<Mover> targets) {
        this.targets = new ArrayList<>(targets);
    }
}