package mv3302.sensor;

import java.awt.geom.Point2D;
import java.util.HashSet;
import java.util.Set;
import simkit.SimEntityBase;
import simkit.smd.Mover;
import simkit.smd.Sensor;

/**
 * An abstract class representing a simple sensor. It extends the SimEntityBase
 * class and implements the Sensor interface. Provides common functionality for
 * sensors that detect and track movers. Subclasses need to implement specific
 * detection algorithms.
 *
 * @author dansl
 */
public abstract class AbstractSimpleSensor extends SimEntityBase implements Sensor {

    private Mover mover;

    private double maxRange;

    protected Set<Mover> contacts;

    /**
     * Constructs a new AbstractSimpleSensor with an empty contacts set.
     */
    public AbstractSimpleSensor() {
        contacts = new HashSet<>();
    }

    /**
     * Constructs a new AbstractSimpleSensor with the specified mover and
     * maximum range. The contacts set is initialized as an empty set.
     *
     * @param mover the mover associated with the sensor
     * @param maxRange the maximum range of the sensor
     */
    public AbstractSimpleSensor(Mover mover, double maxRange) {
        this();
        setMover(mover);
        setMaxRange(maxRange);
    }

    /**
     * Resets the sensor to its initial state. Clears the contacts set.
     */
    @Override
    public void reset() {
        super.reset();
        contacts.clear();
    }

    /**
     * Runs the sensor and fires a property change event for the contacts set.
     */
    public void doRun() {
        firePropertyChange("contacts", getContacts());
    }

    /**
     * Performs detection of a target mover. Adds the target mover to the
     * contacts set and fires property change events for contacts and detection.
     *
     * @param target the mover to be detected
     */
    public void doDetection(Mover target) {
        Set<Mover> oldContacts = getContacts();
        contacts.add(target);
        firePropertyChange("contacts", oldContacts, getContacts());

        firePropertyChange("detection", target);
    }

    /**
     * Removes a target mover from the contacts set. Fires property change
     * events for contacts and undetection.
     *
     * @param target the mover to be undetected
     */
    public void doUndetection(Mover target) {
        Set<Mover> oldContacts = getContacts();
        contacts.remove(target);
        firePropertyChange("contacts", oldContacts, getContacts());

        firePropertyChange("undetection", target);
    }

    /**
     * Initiates the start move action for the specified mover. Delays the
     * action for the current simulation time.
     *
     * @param mover the mover to start moving
     */
    public void doStartMove(Mover mover) {
        waitDelay("StartMove", 0.0, this);
    }

    /**
     * Initiates the stop action for the specified mover. Delays the action for
     * the current simulation time.
     *
     * @param mover the mover to stop
     */
    public void doStop(Mover mover) {
        waitDelay("Stop", 0.0, this);
    }

    /**
     * Initiates the stop action for the specified mover. Delays the action for
     * the current simulation time.
     *
     * @param mover the mover to stop
     */
    public void setMover(Mover mover) {
        this.mover = mover;
        this.mover.addSimEventListener(this);
    }

    /**
     * Returns the maximum range of the sensor.
     *
     * @return the maxRange
     */
    @Override
    public double getMaxRange() {
        return maxRange;
    }

    /**
     * Sets the maximum range of the sensor.
     *
     * @param maxRange the maxRange to set
     */
    public void setMaxRange(double maxRange) {
        if (maxRange < 0.0) {
            throw new IllegalArgumentException("maxRange must be ≥ 0.0: " + maxRange);
        }
        this.maxRange = maxRange;
    }

    /**
     * Sets the contacts of the sensor.
     *
     * @return the contacts
     */
    @Override
    public Set<Mover> getContacts() {
        return new HashSet<>(contacts);
    }

    /**
     * Returns the mover associated with the sensor.
     *
     * @return the mover
     */
    @Override
    public Mover getMover() {
        return mover;
    }

    /**
     * Returns current location the mover associated with the sensor.
     *
     * @return the mover.currentLocation
     */
    @Override
    public Point2D getCurrentLocation() {
        return mover.getCurrentLocation();
    }

    /**
     * Returns velocity of the mover associated with the sensor.
     *
     * @return the movers velocity
     */
    @Override
    public Point2D getVelocity() {
        return mover.getVelocity();
    }

    /**
     * empty method for implementation
     */
    @Override
    public void doStartMove(Sensor sensor) {
    }

    /**
     * empty method for implementation
     */
    @Override
    public void doStop(Sensor sensor) {
    }

    /**
     * Returns a string representation of the AbstractSimpleSensor object. This
     * method is overridden from the superclass to provide a customized string
     * representation.
     *
     * @return a string representation of the AbstractSimpleSensor object
     */
    public String paramString() {
        return super.toString();
    }

    /**
     * Returns a string representation of the AbstractSimpleSensor object. The
     * string format includes the name, current location, velocity, and maximum
     * range of the sensor.
     *
     * @return a string representation of the AbstractSimpleSensor object
     */
    @Override
    public String toString() {
        return String.format("%s %s %s %,.4f", getName(), mover.getCurrentLocation(), mover.getVelocity(),
                getMaxRange());
    }

}
